# -*- coding: utf-8 -*-

import pathlib
from unittest import TestCase

from pymugen.fasta.fastaReader import FastaReader
from pymugen.fasta.sequence import Sequence


class TestFastaReader(TestCase):
    def setUp(self) -> None:
        self.static_dir = f"{str(pathlib.Path(__file__).parent.absolute())}/static/"
        self.reader = FastaReader(
            f"{self.static_dir}test.fa.gz",
        )

        return super().setUp()

    def test__set_fasta_file(self):
        filename = f"{self.static_dir}test.fa"

        result = self.reader.fasta_file.name

        self.assertEqual(result, filename)

    def test_get_sequence(self):
        chromosome = "chr3"
        length = 2
        pos = 9
        from_nuc = 9
        to_nuc = 33
        sequence = Sequence("AGCTAGCTA", "GC", "TAGCTAGCTAGCTAGCTAGCTA")

        result = self.reader.sequence(chromosome, pos, from_nuc, to_nuc, length)

        self.assertEqual(result.prefix, sequence.prefix)
        self.assertEqual(result.infix, sequence.infix)
        self.assertEqual(result.suffix, sequence.suffix)


class TestFastaReader1Line(TestCase):
    def setUp(self) -> None:
        self.static_dir = f"{str(pathlib.Path(__file__).parent.absolute())}/static/"
        self.reader = FastaReader(
            f"{self.static_dir}test2.fa.gz",
        )

        return super().setUp()

    def test__set_fasta_file(self):
        filename = f"{self.static_dir}test2.fa"

        result = self.reader.fasta_file.name

        self.assertEqual(result, filename)

    def test__parse_chromosomes(self):
        data = {
            "1": {
                "name": "1",
                "line_length": 29,
                "label_length": 8,
                "index_start": 0,
                "length": 29,
                "labels": ["chr1", "1"],
            },
            "2": {
                "name": "2",
                "line_length": 36,
                "label_length": 8,
                "index_start": 38,
                "length": 36,
                "labels": ["chr2", "2"],
            },
            "3": {
                "name": "3",
                "line_length": 33,
                "label_length": 8,
                "index_start": 83,
                "length": 33,
                "labels": ["chr3", "3"],
            },
        }

        result = self.reader._parse_chromosomes()

        self.assertEqual(result, data)

    def test_get_sequence(self):
        chromosome = "chr3"
        length = 2
        pos = 9
        from_nuc = 9
        to_nuc = 33
        sequence = Sequence("AGCTAGCTA", "GC", "TAGCTAGCTAGCTAGCTAGCTA")

        result = self.reader.sequence(chromosome, pos, from_nuc, to_nuc, length)

        self.assertEqual(result.prefix, sequence.prefix)
        self.assertEqual(result.infix, sequence.infix)
        self.assertEqual(result.suffix, sequence.suffix)



# -*- coding: utf-8 -*-

from unittest import TestCase

from pymugen.utils.genomics import generate_dict_values


class TestGenomics(TestCase):
    def test_generate_dict_values(self):
        symbols = ["q", "w", "e", "r"]
        mapping = {"A": "q", "C": "w", "G": "e", "T": "r"}

        result = generate_dict_values(symbols)

        self.assertEqual(mapping, result)



# -*- coding: utf-8 -*-

from unittest import TestCase

from pymugen.fasta.sequence import Sequence


class TestMutationParser(TestCase):
    def setUp(self) -> None:
        self.sequence = Sequence("aaa", "bbb", "ccc", "ddd")
        return super().setUp()

    def test_from_string(self):
        sequence_string = str(self.sequence)

        result = Sequence.from_string(sequence_string)

        self.assertEqual(result.prefix, self.sequence.prefix)
        self.assertEqual(result.infix, self.sequence.infix)
        self.assertEqual(result.suffix, self.sequence.suffix)
        self.assertEqual(result.annotations, self.sequence.annotations)
    
    def test_from_string_no_annotation(self):
        sequence = Sequence("aaa", "bbb", "ccc")
        sequence_string = str(sequence)

        result = Sequence.from_string(sequence_string)

        self.assertEqual(result.prefix, sequence.prefix)
        self.assertEqual(result.infix, sequence.infix)
        self.assertEqual(result.suffix, sequence.suffix)
        self.assertEqual(result.annotations, sequence.annotations)
    
    def test_from_string_chromosome(self):
        sequence = Sequence("aaa", "bbb", "ccc", "ddd", "chr1")
        sequence_string = str(sequence)

        result = Sequence.from_string(sequence_string)

        self.assertEqual(result.prefix, sequence.prefix)
        self.assertEqual(result.infix, sequence.infix)
        self.assertEqual(result.suffix, sequence.suffix)
        self.assertEqual(result.annotations, sequence.annotations)

    def test_from_string_no_annotation_chromosome(self):
        sequence = Sequence("aaa", "bbb", "ccc", "", "chr1")
        sequence_string = str(sequence)

        result = Sequence.from_string(sequence_string)

        self.assertEqual(result.prefix, sequence.prefix)
        self.assertEqual(result.infix, sequence.infix)
        self.assertEqual(result.suffix, sequence.suffix)
        self.assertEqual(result.annotations, sequence.annotations)

    def test__sequence_part_str(self):
        sequence_part = "aaa"
        sequence_part_str = f"a-a-a"

        result = self.sequence._sequence_part_str(sequence_part)

        self.assertEqual(result, sequence_part_str)

    def test___str__(self):
        sequence_str = f"""ddd|a-a-a,b-b-b,c-c-c"""

        result = str(self.sequence)

        self.assertEqual(result, sequence_str)

    def test___str__no_annotations(self):
        sequence = Sequence("aaa", "bbb", "ccc")
        sequence_str = f"""a-a-a,b-b-b,c-c-c"""

        result = str(sequence)

        self.assertEqual(result, sequence_str)

    def test___str__chromosme(self):
        sequence = Sequence("aaa", "bbb", "ccc", "ddd", "chr1")
        sequence_str = f"""chr1/ddd|a-a-a,b-b-b,c-c-c"""

        result = str(sequence)

        self.assertEqual(result, sequence_str)

    def test___getitem__1(self):
        sequence_symbols = "aaa"

        result = self.sequence[0]

        self.assertEqual(result, sequence_symbols)

    def test___getitem__2(self):
        sequence_symbols = "bbbccc"

        result = self.sequence[1:]

        self.assertEqual(result, sequence_symbols)


    def test___getitem__3(self):
        sequence_symbols = "aaabbb"

        result = self.sequence[:2]

        self.assertEqual(result, sequence_symbols)

    def test___getitem__4(self):
        sequence_symbols = "bbb"

        result = self.sequence[1:2]

        self.assertEqual(result, sequence_symbols)

    def test___getitem__5(self):
        sequence_symbols = "aaabbbccc"

        result = self.sequence[:]

        self.assertEqual(result, sequence_symbols)

    def test___getitem__6(self):
        error = False

        try:
            self.sequence[5]
        except:
            error = True

        self.assertTrue(error)

    def test___getitem__7(self):
        error = False

        try:
            self.sequence[-1]
        except:
            error = True

        self.assertTrue(error)



# -*- coding: utf-8 -*-

import pathlib
from unittest import TestCase

from pymugen.fasta.sequence import Sequence
from pymugen.transformers.mutationTransformer import MutationTransformer


class TestMutationParser(TestCase):
    def setUp(self) -> None:
        self.static_dir = f"{str(pathlib.Path(__file__).parent.absolute())}/static/"
        self.parser = MutationTransformer(
            f"{self.static_dir}vcfTestTransformers.vcf",
            f"{self.static_dir}test.fa.gz",
        )

        return super().setUp()

    def test_method_insertion(self):
        sequence = Sequence("ACGT", "", "ACGT")
        mutation = "ACGT"
        sequence_transformed = Sequence(
            "qwer",
            "tyui",
            "zxcv",
            ""
        )

        result = self.parser.method(sequence, mutation)

        self.assertEqual(result.prefix, sequence_transformed.prefix)
        self.assertEqual(result.infix, sequence_transformed.infix)
        self.assertEqual(result.suffix, sequence_transformed.suffix)
        self.assertEqual(result.annotations, sequence_transformed.annotations)

    def test_method_delete(self):
        sequence = Sequence("ACGT", "ACGT", "ACGT")
        mutation = ""
        sequence_transformed = Sequence(
            "qwer",
            "ghjk",
            "zxcv",
            "ACGT"
        )

        result = self.parser.method(sequence, mutation)

        self.assertEqual(result.prefix, sequence_transformed.prefix)
        self.assertEqual(result.infix, sequence_transformed.infix)
        self.assertEqual(result.suffix, sequence_transformed.suffix)
        self.assertEqual(result.annotations, sequence_transformed.annotations)

    def test_method_replace(self):
        sequence = Sequence("ACGT", "ACGT", "ACGT")
        mutation = "TGCA"
        sequence_transformed = Sequence(
            "qwer",
            "asdf",
            "zxcv",
            "ACGT"
        )

        result = self.parser.method(sequence, mutation)

        self.assertEqual(result.prefix, sequence_transformed.prefix)
        self.assertEqual(result.infix, sequence_transformed.infix)
        self.assertEqual(result.suffix, sequence_transformed.suffix)
        self.assertEqual(result.annotations, sequence_transformed.annotations)

    def test_method_mix(self):
        sequence = Sequence("ACGT", "AGCT", "ACGT")
        mutation = "GTTCAC"
        sequence_transformed = Sequence(
            "qwer",
            "uadtf",
            "zxcv",
            "AGCT"
        )

        result = self.parser.method(sequence, mutation)

        self.assertEqual(result.prefix, sequence_transformed.prefix)
        self.assertEqual(result.infix, sequence_transformed.infix)
        self.assertEqual(result.suffix, sequence_transformed.suffix)
        self.assertEqual(result.annotations, sequence_transformed.annotations)



# -*- coding: utf-8 -*-

import pathlib
from unittest import TestCase
from unittest.mock import mock_open, patch

from pymugen.fasta.sequence import Sequence
from pymugen.transformers.extendedTransformer import ExtendedTransformer


class TestExtendedParserVcf_ParserVcf(TestCase):
    def setUp(self) -> None:
        self.static_dir = f"{str(pathlib.Path(__file__).parent.absolute())}/static/"
        self.parser = ExtendedTransformer(
            f"{self.static_dir}vcfTestTransformers.vcf",
            f"{self.static_dir}test.fa.gz",
        )

        return super().setUp()

    def test_name(self):
        name = "extended"

        result = self.parser.name

        self.assertEqual(result, name)

    def test_filename(self):
        filename = "transformed_extended_data.pvcf"

        result = self.parser.filename

        self.assertEqual(result, filename)

    def test_method(self):
        sequence = Sequence("ACGT", "ACGT", "ACGT")
        mutation = "TGCA"
        sequence_transformed = Sequence("qwer", "fdsa", "zxcv", "ACGT")

        result = self.parser.method(sequence, mutation)

        self.assertEqual(result.prefix, sequence_transformed.prefix)
        self.assertEqual(result.infix, sequence_transformed.infix)
        self.assertEqual(result.suffix, sequence_transformed.suffix)
        self.assertEqual(result.annotations, sequence_transformed.annotations)

    def test_generate_sequences(self):
        sequences = [
            "G|,a,x-z-v-c-x",
            "G|r-e-w-q-r,f-a,x-z-v-c-x",
            "T|q-r-e-w-q,s,c",
            "T|,a,c-z-x-v-c",
            "T|w-r-e-q-w,f-a,c-z-x-v-c",
            "T|w-r-e-q-w,s,c-z-x-v-c",
        ]
        with patch("builtins.open", mock_open(read_data="data")):
            result = self.parser.generate_sequences("")

        self.assertEqual(result, sequences)

    def test_generate_sequences_chromosome(self):
        sequences = [
            "chr1/G|,a,x-z-v-c-x",
            "chr1/G|r-e-w-q-r,f-a,x-z-v-c-x",
            "chr1/T|q-r-e-w-q,s,c",
            "chr2/T|,a,c-z-x-v-c",
            "chr2/T|w-r-e-q-w,f-a,c-z-x-v-c",
            "chr2/T|w-r-e-q-w,s,c-z-x-v-c",
        ]
        with patch("builtins.open", mock_open(read_data="data")):
            result = self.parser.generate_sequences("", chromosome=True)

        self.assertEqual(result, sequences)

    def test_generate_sequences_original(self):
        sequences = [
            (',G,C-A-T-G-C', 'G|,a,x-z-v-c-x'),
            ('T-G-C-A-T,G,C-A-T-G-C', 'G|r-e-w-q-r,f-a,x-z-v-c-x'),
            ('A-T-G-C-A,T,G', 'T|q-r-e-w-q,s,c'),
            (',T,G-A-C-T-G', 'T|,a,c-z-x-v-c'),
            ('C-T-G-A-C,T,G-A-C-T-G', 'T|w-r-e-q-w,f-a,c-z-x-v-c'),
            ('C-T-G-A-C,T,G-A-C-T-G', 'T|w-r-e-q-w,s,c-z-x-v-c')
        ]
        with patch("builtins.open", mock_open(read_data="data")):
            result = self.parser.generate_sequences("", original=True)

        self.assertEqual(result, sequences)

    def test_retrive_sequence(self):
        sequence = "chr1/G|,a,x-z-v-c-x"
        sequence_transformed = Sequence("", "a", "xzvcx", "G", "chr1")

        result = ExtendedTransformer.retrive_sequence(sequence)

        self.assertEqual(result.prefix, sequence_transformed.prefix)
        self.assertEqual(result.infix, sequence_transformed.infix)
        self.assertEqual(result.suffix, sequence_transformed.suffix)
        self.assertEqual(result.annotations, sequence_transformed.annotations)
        self.assertEqual(result.chromosome, sequence_transformed.chromosome)

    def test_retrive_sequence_original(self):
        sequence = "chr1/,G,C-A-T-G-C\nchr1/G|,a,x-z-v-c-x"
        sequence_transformed = Sequence("", "a", "xzvcx", "G", "chr1")
        sequence_original_transformed = Sequence("", "G", "CATGC", "", "chr1")

        result = ExtendedTransformer.retrive_sequence(sequence)

        self.assertEqual(result[0].prefix, sequence_original_transformed.prefix)
        self.assertEqual(result[0].infix, sequence_original_transformed.infix)
        self.assertEqual(result[0].suffix, sequence_original_transformed.suffix)
        self.assertEqual(result[0].annotations, sequence_original_transformed.annotations)
        self.assertEqual(result[0].chromosome, sequence_original_transformed.chromosome)
        self.assertEqual(result[1].prefix, sequence_transformed.prefix)
        self.assertEqual(result[1].infix, sequence_transformed.infix)
        self.assertEqual(result[1].suffix, sequence_transformed.suffix)
        self.assertEqual(result[1].annotations, sequence_transformed.annotations)
        self.assertEqual(result[1].chromosome, sequence_transformed.chromosome)



# -*- coding: utf-8 -*-

from typing import Union

from pymugen.fasta.sequence import Sequence
from pymugen.transformers.transformer import Transformer
from pymugen.utils.genomics import generate_dict_values

PREFIX_SYMBOLS = ["q", "w", "e", "r"]
MUTATIONS_SYMBOLS = ["a", "s", "d", "f"]
SUFFIX_SYMBOLS = ["z", "x", "c", "v"]


class ExtendedTransformer(Transformer):
    """The extended transformer gets a sequence and transforms it mapping each character
    of the sequence to another depending if the character is in the  prefix, infix or
    suffix.

    Each mapping is defined as:

    - Prefix mapping:
        - A -> q
        - C -> w
        - G -> e
        - T -> r
    - Infix mapping:
        - A -> a
        - C -> s
        - G -> d
        - T -> f
    - Suffix mapping:
        - A -> z
        - C -> x
        - G -> c
        - T -> v

    So, for instance, for a given sequence `Sequence("ACGT", "ACGT", "ACGT")` and a
    mutation `"TGCA"` the  parsers changes it to:

    ```python
        Sequence("qwer", "fdsa", "zxcv", "ACGT")
    ```

    The transformed sequence is named the **extended** sequence.

    Parameters
    ----------
    vcf_path: str
        Path of the vcf file.
    fasta_path: str
        Path of the fasta file.
    """

    prefix_map: dict = generate_dict_values(PREFIX_SYMBOLS)
    """Mapping between nucleotides and the prefix symbols."""

    mutations_map: dict = generate_dict_values(MUTATIONS_SYMBOLS)
    """Mapping between nucleotides and the mutation symbols."""

    suffix_map: dict = generate_dict_values(SUFFIX_SYMBOLS)
    """Mapping between nucleotides and the suffix symbols."""

    name: str = "extended"

    @classmethod
    def method(cls, sequence: Union[tuple, list], mutation: str) -> Sequence:
        """Transforms a sequence into a extended representation.

        Parameters
        ----------
        sequence: tuple
            Sequence.
        mutation: str
            Mutation sequence.

        Returns
        -------
        Transformed sequence.
        """

        left = "".join([cls.prefix_map[nucletid.upper()] for nucletid in sequence[0]])
        middle = "".join([cls.mutations_map[nucletid.upper()] for nucletid in mutation])
        right = "".join([cls.suffix_map[nucletid.upper()] for nucletid in sequence[2]])

        return Sequence(left, middle, right, sequence[1], sequence.chromosome)



# -*- coding: utf-8 -*-

import gzip
import os
import shutil

from pymugen.fasta.chromosome import Chromosome
from pymugen.fasta.sequence import Sequence


class FastaReader(object):
    """Gets a FASTA file and parse it, getting information about the chromosomes.

    After parse the fasta file, the class will have a dictionary with information per
    chromosome. Each value is a `pymugen.fasta.chromosome.Chromosome` class.

    For example, after analize a Fasta file with two chromosomes, the dictionary will be:

    ```python
        chromosomes = {
            'chr1': Chromosome(...),
            'chr2': Chromosome(...)
        }
    ```

    This class allow to get sequences from the fasta file from chromosomes using
    `get_sequence` method.

    To acces to the information of one chromosome we can use array acces way:

    ```python
        fr = FastaReader(...)

        chr = fr['chr1']

        sequence = chr.sequence(...)
        sequence_string = chr[2:9]
    ```

    Parameters
    ----------
    fasta_path : str
        Path of the fasta file.
    """

    chromosomes: dict = {}
    """The dictionary that contains all chromosomes data of the FASTA file, Each key is
    the identifier of each chromosome."""

    _labels: dict = {}
    """Mapping between labels and chromosome identifier."""

    _chromosomes_list: list = []
    """List of the identifiers of all the chromosomes of the FASTA file."""

    fasta_filename: str = None
    """Name of the fasta file"""

    def __init__(self, fasta_path: str):
        self.fasta_filename = fasta_path.replace(".gz", "")
        self._set_fasta_file(fasta_path)
        self._parse_chromosomes()

    def _set_fasta_file(self, fasta_path: str):
        """Gets the fasta file from gz file and unzip the file to get the complete fasta
        file. If there is a unziped file with the same name without .gz extension, the
        class use that file instead of unzip the gz one.

        Parameters
        ----------
        fasta_path : str
            Path of the fasta file.
        """
        if not os.path.isfile(self.fasta_filename):
            self.fasta_file = gzip.open(fasta_path, "r")

            with open(self.fasta_filename, "wb") as f_out:
                shutil.copyfileobj(self.fasta_file, f_out)

        self.fasta_file = open(self.fasta_filename, "r")

    def _parse_chromosomes(self) -> dict:
        """Create all Cromosmes of he FASTA file and append it to a dictionary with the
        names of chromosomes as keys.

        For example, after analize a Fasta file with two chromosomes, the dictionary
        will be:

        ```python
            chromosomes = {
                'chr1': Chromosome(),
                'chr2': Chromosome()
            }
        ```

        Returns
        -------
        Chromosomes.
        """
        self._chromosomes_list = []
        self.chromosomes = {}

        lines = 0
        index = 0
        length = 0
        result = {}
        current_chromosome = False
        self.fasta_file.seek(0, 0)
        for i in self.fasta_file:
            lines += 1
            if i.startswith(">"):
                if current_chromosome:
                    result[current_chromosome]["length"] = length
                length = 0
                label_length = len(i)
                labels = i.rstrip().replace(">", "").split()
                current_chromosome = labels[len(labels) - 1]

                self._chromosomes_list.append(current_chromosome)
                result[current_chromosome] = {
                    "name": current_chromosome,
                    "line_length": False,
                    "label_length": label_length,
                    "index_start": index,
                    "length": False,
                    "labels": labels,
                }
                for label in labels:
                    self._labels[label] = current_chromosome
                is_populated = False
                index += len(i)
                continue

            line_length = len(i.rstrip())
            index += len(i)
            length += line_length

            if is_populated:
                continue

            is_populated = True
            result[current_chromosome]["line_length"] = line_length

        result[current_chromosome]["length"] = length

        for i in result:
            chromosome = Chromosome(
                self.fasta_file,
                name=i,
                line_length=result[i]["line_length"],
                label_length=result[i]["label_length"],
                index_start=result[i]["index_start"],
                length=result[i]["length"],
                labels=result[i]["labels"],
            )

            self.chromosomes[i] = chromosome

        return result

    def __getitem__(self, key):
        assert isinstance(key, str)

        return self.chromosomes[self._labels[key]]

    def sequence(
        self,
        chromosome: str,
        pos: int,
        from_nuc: int,
        to_nuc: int,
        length: int = 1,
    ) -> Sequence:
        """Gets a sequence from the fasta file by the position of a nucleotide on a
        chromosome, with a specified prefix length and suffix length.

        Parameters
        ----------
        chromosome : str
            The chromosome where the sequence is going to be obtained.
        pos : int
            Position of the nucleotide on the chromosome.
        from_nuc : int
            Length of the prefix of the sequence.
        to_nuc : int
            Length of the suffix of the sequence.
        length : int = 1
            Length of the infix.

        Returns
        -------
        The sequence.
        """
        return self[chromosome].sequence(pos, from_nuc, to_nuc, length=length)



from typing import Union

import Levenshtein
from pymugen.fasta.sequence import Sequence
from pymugen.transformers.extendedTransformer import ExtendedTransformer
from pymugen.utils.genomics import generate_dict_values

MUTATIONS_INSERTION_SYMBOLS = ["t", "y", "u", "i"]
MUTATIONS_EARSED_SYMBOLS = ["g", "h", "j", "k"]
MUTATIONS_SUBSITUTION_SYMBOLS = ["a", "s", "d", "f"]

INSERT = "insert"
DELETE = "delete"
REPLACE = "replace"

MUTATION_TYPES = [
    (INSERT, MUTATIONS_INSERTION_SYMBOLS),
    (DELETE, MUTATIONS_EARSED_SYMBOLS),
    (REPLACE, MUTATIONS_SUBSITUTION_SYMBOLS),
]


class MutationTransformer(ExtendedTransformer):
    """The **mutation transformer** is similar to the **extended transformer**, the
    difference is that the result infix has a different symbol depending on how changed
    is the infix reference and the mutation.
    
    In the mutation transformer, each type of modification between the reference
    and the mutation has a different symbol.

    This means that we can get three types of mutations:

    - **Insertion**
    - **Erase**
    - **Substitution**

    So, if we have "A", "C", "G" and "T" symbols for the infix, each symbol has three
    possible mutations:

        A -> ["A_insertion", "A_earsed", "A_substitution"]
        C -> ["C_insertion", "C_earsed", "C_substitution"]
        G -> ["G_insertion", "G_earsed", "G_substitution"]
        T -> ["T_insertion", "T_earsed", "T_substitution"]

    In this case, this method differentiates between prefix symbols and suffix symbols
    (as the extended transformer), so the mappings of each symbol in each position is:

    - **Prefix**
        - A -> q
        - C -> w
        - G -> e
        - T -> r
    - **Infix**
        - **Insertion**
            - A -> t
            - C -> y
            - G -> u
            - T -> i
        - **Erase**
            - A -> g
            - C -> h
            - G -> j
            - T -> k
        - **Substitution**
            - A -> a
            - C -> s
            - G -> d
            - T -> f
    - **Suffix**
        - A -> z
        - C -> x
        - G -> c
        - T -> v
    
    So, for instance, for a given sequence `Sequence("AA", "AC", "CC")` and a
    mutation `"TTGT"` the  parsers changes it to:

    ```python
        Sequence("qq", "iias", "xx", "AC")
    ```

    The transformed sequence is named the **mutation** sequence.

    Parameters
    ----------
    vcf_path: str
        Path of the vcf file.
    fasta_path: str
        Path of the fasta file.
    """

    name: str = "mutation"

    mutations_map: dict = {
        operation: generate_dict_values(symbols)
        for operation, symbols in MUTATION_TYPES
    }

    @classmethod
    def method(cls, sequence: Union[tuple, list], mutation: str) -> Sequence:
        """Transforms a sequence into a extended representation.

        Parameters
        ----------
        sequence : tuple
            Sequence.
        mutation : str
            Mutation sequence.

        Returns
        -------
        Transformed sequence.
        """
        infix = sequence[1]

        operations = Levenshtein.editops(infix, mutation)

        mutation_type_sequence = []
        for operation in operations:
            mutation_type = operation[0]

            reference_sequence = infix
            index = 1
            if mutation_type == INSERT:
                reference_sequence = mutation
                index = 2

            reference_symbol = reference_sequence[operation[index]]

            mutation_type_symbol = cls.mutations_map[mutation_type][reference_symbol]

            mutation_type_sequence.append(mutation_type_symbol)

        left = "".join([cls.prefix_map[nucletid.upper()] for nucletid in sequence[0]])
        right = "".join([cls.suffix_map[nucletid.upper()] for nucletid in sequence[2]])
        mutation_type_sequence = "".join(mutation_type_sequence)

        return Sequence(left, mutation_type_sequence, right, infix, sequence.chromosome)



from typing import Union

NUCLETODIES: list = ["A", "C", "G", "T"]
""" List of nucleotides """


def generate_dict_values(
    lst: Union[tuple, list], reference: Union[tuple, list] = NUCLETODIES
) -> dict:
    """Function that creates a dcitionary that represents a map between a list of
    symbols and the reference nucleotides (A, C, G and T for example)

    For example, if our list is:

        ['q', 'w', 'e', 'r']

    the function will return:

        {
            'A': 'q'
            'C': 'w'
            'G': 'e'
            'T': 'r'
        }

    Parameters
    ----------
    lst: list, tuple
        List of symbols
    reference: list, tuple = ["A", "C", "G", "T"]
        List of reference symbols

    Returns
    -------
    Dictionary with the mapping
    """
    return {key: value for (key, value) in zip(reference, lst)}



from typing import TextIO

from pymugen.fasta.sequence import Sequence


class Chromosome(object):
    """Class that represents a chromosome.

    Using this class a sequence can be obtained as list, for example, if we want to get
    a sequence that starts at position 123 and finish at 456, we can get it using:

    ```python
        chromosome[123:457]
    ```

    If we want to get a prefix from a position, for example, 7 symbols:

    ```python
        chromosome[123:-7]
    ```

    Parameters
    ----------
    fasta_file: TextIO
        Fasta file.
    name: str
        Name of the chromosome.
    line_length: int
        Line length of the chromosome.
    label_length: int
        Length of the first line of the chromosome.
    index_start: int
        Index where the chrosomosme starts on the FASTA file.
    length: int
        Length of the chromosome.
    """

    def __init__(
        self,
        fasta_file: TextIO,
        name: str,
        line_length: int,
        label_length: int,
        index_start: int,
        length: int,
        labels: list = False,
    ) -> None:
        self.name = name
        self.fasta_file = fasta_file
        self.line_length = line_length
        self.label_length = label_length
        self.index_start = index_start
        self.length = length
        self.labels = labels

    def sequence(
        self,
        pos: int,
        from_nuc: int,
        to_nuc: int,
        length: int = 1,
    ) -> Sequence:
        """Gets a sequence from the fasta file by the position of a nucleotide on a
        chromosome, with a specified prefix length and suffix length.

        Parameters
        ----------
        chromosome : str
            The chromosome where the sequence is going to be obtained.
        pos : int
            Position of the nucleotide on the chromosome.
        from_nuc : int
            Length of the prefix of the sequence.
        to_nuc : int
            Length of the suffix of the sequence.
        length : int = 1
            Length of the infix.

        Returns
        -------
        The sequence divided in (prefix, nucleotide, suffix).
        """
        pref = self[pos:-from_nuc]
        nucleotide = self[pos : pos + length]
        suff = self[pos + length : pos + length + to_nuc]

        return Sequence(pref, nucleotide, suff)

    def _get_nucleotide_index(self, pos: int) -> int:
        """Gets the index of a nucleotide by its position in a chromosome.

        Parameters
        ----------
        pos : int
            Index of the nucletoide on the chromosome.

        IndexError
            When index is greater tha chromsome length or lower than 0.

        Returns
        -------
        Index of the nucleotide on the fasta file.
        """
        length = self.length - 1

        if pos > length or pos < 0:
            raise IndexError(f"Invalid index, must be in the interval {0}-{length}")

        # It's necessary to taking into account that seek method counts new lines as a
        # character, that's why we add new line characters ('\n')
        num_new_lines = int(pos / self.line_length)
        index_start = self.index_start
        label_length = self.label_length

        # Get the position of the nucleotid on the file (1 char is one byte, that's why
        # we use seek)
        return pos + index_start + label_length + num_new_lines

    def _get_from_interval(self, starts: int, length: int) -> str:
        """Returns a sequence that starts at a given index of the fasta file and has a
        given length.

        Parameters
        ----------
        length : int
            Length of the sequence.
        starts : int
            Index of the fasta file where the sequence starts.

        Raises
        ------
        IndexError
            When the start position is wrong or invalid.

        Returns
        -------
        The sequence.
        """
        self.fasta_file.seek(starts, 0)
        sequence = ""

        while length != 0:
            searched_sequence = self.fasta_file.read(length).split("\n")
            # Number of newline characters present on the sequence searched
            length = len(searched_sequence) - 1
            sequence += "".join(searched_sequence)

        return sequence.upper()

    def __getitem__(self, key):
        if isinstance(key, (int)):
            key = slice(key, key + 1, None)

        start = key.start or 0

        if start < 0:
            raise IndexError()

        stop = key.stop or self.length
        if stop < 0:
            stop, start = start, start + stop
            if start < 0:
                start = 0

        if start < 0:
            start = 0
        if stop > self.length:
            stop = self.length

        length = stop - start
        return self._get_from_interval(self._get_nucleotide_index(start), length)

    def __ln__(self):
        return self.length



# -*- coding: utf-8 -*-

import pathlib
from unittest import TestCase

from pymugen.fasta.chromosome import Chromosome
from pymugen.fasta.sequence import Sequence


class TestChromosomeCHR1(TestCase):
    def setUp(self) -> None:
        self.static_dir = f"{str(pathlib.Path(__file__).parent.absolute())}/static/"
        self.sequence = "GCATGCATGCATGCATGCATGCATGCATG"
        self.chromosome = Chromosome(
            open(f"{self.static_dir}test.fa", "r"),
            name="chr1",
            line_length=12,
            label_length=6,
            index_start=0,
            length=29,
        )

        return super().setUp()

    def test___getitem__(self):
        sequence = self.sequence[0]

        data = self.chromosome[0]

        self.assertEqual(data, sequence)

    def test___getitem___no_start(self):
        sequence = self.sequence[:24]

        data = self.chromosome[:24]

        self.assertEqual(data, sequence)

    def test___getitem___no_end(self):
        sequence = self.sequence[4:]

        data = self.chromosome[4:]

        self.assertEqual(data, sequence)

    def test___getitem___all(self):
        sequence = self.sequence[:]

        data = self.chromosome[:]

        self.assertEqual(data, sequence)

    def test__get_prefix_at_start(self):
        pos = 5
        length = 7
        prefix = self.sequence[0:5]

        data = self.chromosome[pos:-length]

        self.assertEqual(data, prefix)

    def test__get_prefix(self):
        pos = 13
        length = 7
        prefix = self.sequence[pos - length : pos]

        data = self.chromosome[pos:-length]

        self.assertEqual(data, prefix)

    def test__get_prefix_at_end(self):
        pos = 28
        length = 7
        prefix = self.sequence[pos - length : pos]

        data = self.chromosome[pos:-length]

        self.assertEqual(data, prefix)

    def test__get_nucleotide_index_pos_greater_length(self):
        pos = 1000
        error = False

        try:
            self.chromosome._get_nucleotide_index(pos)
        except IndexError:
            error = True

        self.assertTrue(error)

    def test__get_nucleotide_index_pos_lower_0(self):
        pos = -1
        error = False

        try:
            self.chromosome._get_nucleotide_index(pos)
        except IndexError:
            error = True

        self.assertTrue(error)

    def test__get_nucleotide_index(self):
        pos = 24
        index = 32

        result = self.chromosome._get_nucleotide_index(pos)

        self.assertEqual(result, index)

    def test__get_from_interval(self):
        index = 6
        length = 29
        sequence = "GCATGCATGCATGCATGCATGCATGCATG"

        result = self.chromosome._get_from_interval(index, length)

        self.assertEqual(result, sequence)


class TestChromosomeCHR2(TestCase):
    def setUp(self) -> None:
        self.static_dir = f"{str(pathlib.Path(__file__).parent.absolute())}/static/"
        self.sequence = "TGACTGACTGACTGACTGACTGACTGACTGACTGAC"
        self.chromosome = Chromosome(
            open(f"{self.static_dir}test.fa", "r"),
            name="chr2",
            line_length=12,
            label_length=6,
            index_start=38,
            length=36,
        )

        return super().setUp()

    def test___getitem__(self):
        sequence = self.sequence[0]

        data = self.chromosome[0]

        self.assertEqual(data, sequence)

    def test___getitem___no_start(self):
        sequence = self.sequence[:24]

        data = self.chromosome[:24]

        self.assertEqual(data, sequence)

    def test___getitem___no_end(self):
        sequence = self.sequence[4:]

        data = self.chromosome[4:]

        self.assertEqual(data, sequence)

    def test___getitem___all(self):
        sequence = self.sequence[:]

        data = self.chromosome[:]

        self.assertEqual(data, sequence)

    def test__get_prefix_at_start(self):
        pos = 5
        length = 7
        prefix = self.sequence[0:5]

        data = self.chromosome[pos:-length]

        self.assertEqual(data, prefix)

    def test__get_prefix(self):
        pos = 13
        length = 7
        prefix = self.sequence[pos - length : pos]

        data = self.chromosome[pos:-length]

        self.assertEqual(data, prefix)

    def test__get_prefix_at_end(self):
        pos = 28
        length = 7
        prefix = self.sequence[pos - length : pos]

        data = self.chromosome[pos:-length]

        self.assertEqual(data, prefix)


class TestChromosomeCHR3(TestCase):
    def setUp(self) -> None:
        self.static_dir = f"{str(pathlib.Path(__file__).parent.absolute())}/static/"
        self.sequence = "AGCTAGCTAGCTAGCTAGCTAGCTAGCTAGCTA"
        self.chromosome = Chromosome(
            open(f"{self.static_dir}test.fa", "r"),
            name="chr3",
            line_length=12,
            label_length=6,
            index_start=83,
            length=33,
        )

        return super().setUp()

    def test___getitem__(self):
        sequence = self.sequence[0]

        data = self.chromosome[0]

        self.assertEqual(data, sequence)

    def test___getitem___no_start(self):
        sequence = self.sequence[:24]

        data = self.chromosome[:24]

        self.assertEqual(data, sequence)

    def test___getitem___no_end(self):
        sequence = self.sequence[4:]

        data = self.chromosome[4:]

        self.assertEqual(data, sequence)

    def test___getitem___all(self):
        sequence = self.sequence[:]

        data = self.chromosome[:]

        self.assertEqual(data, sequence)

    def test__get_prefix_at_start(self):
        pos = 5
        length = 7
        prefix = self.sequence[0:5]

        data = self.chromosome[pos:-length]

        self.assertEqual(data, prefix)

    def test__get_prefix(self):
        pos = 13
        length = 7
        prefix = self.sequence[pos - length : pos]

        data = self.chromosome[pos:-length]

        self.assertEqual(data, prefix)

    def test__get_prefix_at_end(self):
        pos = 28
        length = 7
        prefix = self.sequence[pos - length : pos]

        data = self.chromosome[pos:-length]

        self.assertEqual(data, prefix)

    def test_get_sequence(self):
        pos = 9
        from_nuc = 9
        length = 2
        to_nuc = 33
        sequence = Sequence("AGCTAGCTA", "GC", "TAGCTAGCTAGCTAGCTAGCTA")

        result = self.chromosome.sequence(pos, from_nuc, to_nuc, length)

        self.assertEqual(result.prefix, sequence.prefix)
        self.assertEqual(result.infix, sequence.infix)
        self.assertEqual(result.suffix, sequence.suffix)

class TestChromosome1LineCHR2(TestCase):
    def setUp(self) -> None:
        self.static_dir = f"{str(pathlib.Path(__file__).parent.absolute())}/static/"
        self.sequence = "TGACTGACTGACTGACTGACTGACTGACTGACTGAC"
        self.chromosome = Chromosome(
            open(f"{self.static_dir}test2.fa", "r"),
            name="chr2",
            line_length=36,
            label_length=8,
            index_start=38,
            length=36,
        )

        return super().setUp()

    def test___getitem__(self):
        sequence = self.sequence[0]

        data = self.chromosome[0]

        self.assertEqual(data, sequence)

    def test___getitem___no_start(self):
        sequence = self.sequence[:24]

        data = self.chromosome[:24]

        self.assertEqual(data, sequence)

    def test___getitem___no_end(self):
        sequence = self.sequence[4:]

        data = self.chromosome[4:]

        self.assertEqual(data, sequence)

    def test___getitem___all(self):
        sequence = self.sequence[:]

        data = self.chromosome[:]

        self.assertEqual(data, sequence)

    def test__get_prefix_at_start(self):
        pos = 5
        length = 7
        prefix = self.sequence[0:5]

        data = self.chromosome[pos:-length]

        self.assertEqual(data, prefix)

    def test__get_prefix(self):
        pos = 13
        length = 7
        prefix = self.sequence[pos - length : pos]

        data = self.chromosome[pos:-length]

        self.assertEqual(data, prefix)

    def test__get_prefix_at_end(self):
        pos = 28
        length = 7
        prefix = self.sequence[pos - length : pos]

        data = self.chromosome[pos:-length]

        self.assertEqual(data, prefix)



from pymugen.transformers.extendedTransformer import ExtendedTransformer
from pymugen.transformers.mutationTransformer import MutationTransformer
from pymugen.transformers.transformer import Transformer



# -*- coding: utf-8 -*-

from abc import ABC, abstractmethod
from typing import Sequence, Union

from pymugen.fasta.fastaReader import FastaReader
from pymugen.fasta.sequence import Sequence
from vcf import Reader as VcfReader


class Transformer(ABC):
    """Transforms data from a VCF file using a FASTA file.

    The transformer needs to define a method named `method`, which generates a new
    sequence from the original one with different notation.

    For example, if we have the sequence `abbc = Sequence(a, bb, c)` and the mutation
    `aa` this `method` changes it into another representation, for example to
    `smms = Sequence(s, mm, s, bb)`, where the mutation will be added and transformed
    and the reference infix is saved as annotation.

    ```python
        Sequence("s", "mm", "s", "b") = method(Sequence("a", "b", "c"), "bb")
    ```

    Parameters
    ----------
    vcf_path: str
        Path of the vcf file.
    fasta_path: str
        Path of the fasta file.
    """

    fasta_reader: FastaReader = None
    _sequences_separator: str = "\n"

    def __init__(self, vcf_path: str, fasta_path: str):
        self.vcf = VcfReader(open(vcf_path, "r"))
        self.fasta_reader = FastaReader(fasta_path)

    @property
    @abstractmethod
    def name(self) -> str:
        """Name of the transformer."""
        pass

    @property
    def filename(self) -> str:
        """Default filename for the results."""
        return f"transformed_{self.name}_data.pvcf"

    @abstractmethod
    def method(self, sequence: Sequence, mutation: str) -> Sequence:
        """Transforms the sequence with a given mutation to a new sequence with
        different notation.

        For example, if we have the sequence `abbc = Sequence(a, bb, c)` and the
        mutation `aa` this `method` changes it into another representation, for example
        to `smms = Sequence(s, mm, s, bb)`, where the mutation will be added and
        transformed and the reference infix is saved as annotation.

        ```python
            Sequence("s", "mm", "s", "b") = method(Sequence("a", "b", "c"), "bb")
        ```

        Parameters
        ----------
        sequence: Sequence
            Sequence.
        muatation: str
            Mutation.

        Returns
        -------
        Transformed sequence.
        """
        pass

    def generate_sequences(
        self,
        path: str,
        prefix_length: int = 5,
        suffix_length: int = 5,
        chromosome: bool = False,
        filename: str = False,
        original: bool = False
    ):
        """Generates a file with the mutated sequences using the method `method` for
        transform the sequence and the mutation and returns a list of sequence or pairs
        of sequences if the parameter orignal is set to True.

        Parameters
        ----------
        path: str
            Path to store the data.
        prefix_length: int = 5
            Length of the prefix.
        suffix_length: int = 5
            Length of the suffix.
        chromosome: bool = False
            If true add the chromosome where the sequences is from into the file.
        filename: str = default_filename
            Filename of the result file.
        original: bool = True
            Adds the original sequence into the file and changes the result from a list
            of sequences to a list of pairs of sequences, where the first sequence is
            the orignal and the second the transformed

        Returns
        -------
        Transformed sequences from vcf.
        """

        if not filename:
            filename = self.filename

        sequences = []
        with open(f"{path}/{filename}", "w") as transformed_data_file:
            for i in self.vcf:
                sequence = self.fasta_reader.sequence(
                    i.CHROM, i.POS - 1, prefix_length, suffix_length, len(i.REF)
                )

                assert sequence.infix.upper() == i.REF.upper()

                if chromosome:
                    sequence.chromosome = i.CHROM

                transformed_sequence = self.method(sequence, i.ALT[0].sequence)

                result = str(transformed_sequence)
                element = result
                if original:
                    result = f"{sequence}{self._sequences_separator}{result}"
                    element = (str(sequence), element)

                sequences.append(element)
                transformed_data_file.write(result)

        return sequences

    @classmethod
    def retrive_sequence(cls, sequence: str) -> Sequence:
        """Retrive a string sequence into a sequence object.

        Parameters
        ----------
        sequence: str
            Sequence in a string format.

        Returns
        -------
        Sequence or sequences.
        """
        if cls._sequences_separator in sequence:
            original, sequence = sequence.split(cls._sequences_separator)
            return (Sequence.from_string(original), Sequence.from_string(sequence))
        return Sequence.from_string(sequence)



class Sequence(object):
    """Class that represents a genomic sequence. It has three elements, a prefix an
    infix and a suffix. In addition, the sequence has two attributes, one for the
    chromosome where the sequence is in and the other for annotations.

    This sequence can be represented as a string with the next format:

    ```
    chr/annotations|symbol_p_1-symbol_p_n,symbol_m_1-symbol_m_n,symbol_s_1-symbol_n_1
    ```

    The sequence can be accessed as a list of three elements, where each element is one
    part of the examples, for example:

    ```python
    sequence = Sequence("aaa", "bbb", "ccc")

    sequence[1]
    >>> bbb

    sequence[0:2]
    >>> aaabbb

    sequence[:]
    >>> aaabbbccc
    ```

    The literal symbols can be changed by inheriting the class and changing these
    variables:

     - `_prefix_suffix_separator = "|"`
     - `_sequence_part_separator = ","`
     - `_symbol_sequence_separator = "-"`
     - `_chromosome_separator = "/"`
    
    Parameters
    ----------
    prefix: str
        Prefix of the sequence.
    infix: str
        Infix of the sequence.
    suffix: str
        Suffix of the sequence.
    annotations: str = ""
        Annotations of the sequence.
    chromosome: str = ""
        Chromosome of the sequence.
    """

    _prefix_separator: str = "|"
    _sequence_part_separator: str = ","
    _symbol_sequence_separator: str = "-"
    _chromosome_separator: str = "/"

    def __init__(
        self,
        prefix: str,
        infix: str,
        suffix: str,
        annotations: str = "",
        chromosome: str = "",
    ) -> None:
        self.prefix = prefix
        self.infix = infix
        self.suffix = suffix
        self.sequence = (self.prefix, self.infix, self.suffix)

        self.annotations = annotations
        self.chromosome = chromosome
        super().__init__()

    @classmethod
    def from_string(cls, sequence: str):
        """Creates a sequence from a string sequence.
        
        Parameters
        ----------
        sequence: str
            Sequence in a string format.
        
        Returns
        -------
        Sequence object.
        """
        annotations = ""
        if cls._prefix_separator in sequence:
            annotations, sequence = sequence.split(cls._prefix_separator)

        chromosome = ""
        if cls._chromosome_separator in annotations:
            chromosome, annotations = annotations.split(cls._chromosome_separator)
        elif cls._chromosome_separator in sequence:
            chromosome, sequence = sequence.split(cls._chromosome_separator)

        prefix, infix, suffix = sequence.split(cls._sequence_part_separator)

        prefix = prefix.replace(cls._symbol_sequence_separator, "")
        infix = infix.replace(cls._symbol_sequence_separator, "")
        suffix = suffix.replace(cls._symbol_sequence_separator, "")

        return Sequence(prefix, infix, suffix, annotations, chromosome)

    def _sequence_part_str(self, sequence: str) -> str:
        """Joins a sequence using the symbols separator.
        
        Parameters
        ----------
        sequence: str
            Part of a sequence.
        
        Returns
        -------
        Sequence divided by the symbol separator.
        """
        return self._symbol_sequence_separator.join(sequence)

    def __str__(self):
        prefix_transformed = self._sequence_part_str(self.prefix)
        infix_transformed = self._sequence_part_str(self.infix)
        suffix_transformed = self._sequence_part_str(self.suffix)

        sequence = self._sequence_part_separator.join(
            [prefix_transformed, infix_transformed, suffix_transformed]
        )

        prefix = ""
        if self.annotations:
            prefix = f"{self.annotations}{self._prefix_separator}"

        if self.chromosome:
            prefix = f"{self.chromosome}{self._chromosome_separator}{prefix}"

        return f"{prefix}{sequence}"

    def __getitem__(self, key):
        start = None
        stop = None
        if type(key) is int:
            start = key
            stop = key + 1
            key = slice(None, None, None)

        start = start or key.start
        if start is None:
            start = start or 0

        stop = stop or key.stop
        if stop is None:
            stop = stop or 3

        if stop > 3 or start < 0:
            raise ValueError("Bad index, must be between 0 and 2")

        result = ""
        for index in range(start, stop):
            result += self.sequence[index]

        return result
